#include <Node.h>
#include <NodeInfo.h>
#include <Path.h>
#include <stdio.h>
#include <malloc.h>
#include <SymLink.h>
#include <string.h>
#include <Volume.h>
#include <time.h>
#include <Window.h>
#include <PopUpMenu.h>
#include <Menu.h>
#include <MenuItem.h>
#include "basTumbnailView.h"
#include "basApplication.h"
#include "BitmapUtil.h"
#include "basCommon.h"
#include "basBitmapWorker.h"
#include "Colors.h"
#include "basWindow.h"
#include "basLocale.h"
#include "TextDefs.h"
#include "TumbnailCache.h"

void Entry_StringSize( int size, char *out );

basTumbnailWorker::basTumbnailWorker( basTumbnailView *tv )
	: TV(tv), tID(-1)
{
	workList = new BList();
	Semaphore = create_sem(1, "TumbnailWorker_Semaphore");
}

basTumbnailWorker::~basTumbnailWorker()
{
	Stop();
	delete_sem(Semaphore);
	delete workList;
}

void basTumbnailWorker::Stop()
{
	acquire_sem(Semaphore);
	if ( tID >= 0)
	{
		kill_thread(tID);
		tID = -1;
	}

	BMessage *msg;

	while ( (msg = (BMessage*)workList->FirstItem()) )
	{
		workList->RemoveItem(msg);
		delete msg;
	}
	release_sem(Semaphore);
}

void basTumbnailWorker::PostData( BMessage *msg )
{
	acquire_sem(Semaphore);
	workList->AddItem(msg);

	if ( tID == -1 )
	{
		tID = spawn_thread( doWork, "TumbnailWorker", B_NORMAL_PRIORITY, (void*)this );
		resume_thread(tID);
	}

	release_sem(Semaphore);
//	doWork(this);
}

int32 basTumbnailWorker::doWork( void *_worker )
{
	basTumbnailWorker *worker = (basTumbnailWorker*)_worker;
	BMessage *msg;

	acquire_sem(worker->Semaphore);

	while ( (msg = (BMessage*)worker->workList->FirstItem()) )
	{
		if ( ((basApplication*)be_app)->fQuiting )
			return 0;

		entry_ref ref;
		node_ref node;
		char *name;
		ino_t ino;
		dev_t dev;
		msg->FindRef("_ref", &ref);
		msg->FindString("_name", (const char **)&name);
		msg->FindInt32("_dev", &dev);
		msg->FindInt64("_ino", &ino);
		node.device = dev;
		node.node = ino;
		BEntry ent(&ref);
		worker->TV->ShowFolderEaner( &ent, name, ref, node );

		worker->workList->RemoveItem(msg);
		delete msg;
	}

	worker->tID = -1;

	release_sem(worker->Semaphore);

	return 0;
}

basStatusBarView::basStatusBarView( BView *view, BView *status )
	: BView( BRect(0, 0, 300, 300), NULL, B_FOLLOW_ALL_SIDES, B_FRAME_EVENTS | B_NAVIGABLE ),
	View(view), StatusBar(status)
{
	AddChild(View);
	AddChild(StatusBar);
}

void basStatusBarView::FrameResized( float x, float y )
{
	View->MoveTo( 0, 0 );
	View->ResizeTo( x+1, y - B_V_SCROLL_BAR_WIDTH+1 );
	StatusBar->MoveTo( 0, y - B_V_SCROLL_BAR_WIDTH+1);
	StatusBar->ResizeTo( x, B_V_SCROLL_BAR_WIDTH-1);
}

void basStatusBarView::MessageReceived( BMessage *msg )
{
	switch ( msg->what )
	{
		case B_MOUSE_DOWN:
		case B_KEY_DOWN:
			Window()->PostMessage(msg);
			break;
		default:
			BView::MessageReceived(msg);
	}
}

basStatusBar::basStatusBar()
	: BView( BRect(0,0,100,B_V_SCROLL_BAR_WIDTH), NULL, B_FOLLOW_ALL_SIDES, B_WILL_DRAW ),
	Text(0), NextText(0)
{
	SetViewColor( BeBackgroundGrey );
	SetLowColor( BeBackgroundGrey );
}

void basStatusBar::SetLabel( char *text )
{
	if ( Text )
		delete Text;
	Text = new char [ strlen(text)+1 ];
	strcpy(Text, text);
	Invalidate();
}

void basStatusBar::SetNextLabel( char *text )
{
	if ( NextText )
		delete NextText;
	if ( text )
	{
		NextText = new char [ strlen(text)+1 ];
		strcpy(NextText, text);
	}
	else
		NextText = NULL;
	Invalidate();
}

void basStatusBar::Draw( BRect draw_rect )
{
	SetHighColor( BeShadow );
	FillRect( BRect( 0, 0, Bounds().Width(), 0 ) );
	SetHighColor( White );
	FillRect( BRect( 0, 1, Bounds().Width(), 1 ) );
	SetHighColor( BeShadow );
	FillRect( BRect( 0, Bounds().bottom, Bounds().Width(), Bounds().bottom ) );
	if ( Text )
	{
		SetHighColor( Black );
		DrawString( Text, BPoint( 5, Bounds().Height() - 2 ) );
		if ( NextText )
			DrawString( NextText, BPoint( 5 + StringWidth(Text) + 3, Bounds().Height() - 2));
	}
}

#define TUMBNAILSIZE28		(tumbnailSize+28+24)

void Entry_StringSize( BEntry *, off_t *, char * );

void EnshortStringToFit( char *in, char *out, float size, BView *view );
void to_rgb( int32 color, rgb_color &co );
int32 rgb_to( rgb_color &co );

void EnshortStringToFit( char *in, char *out, float size, BView *view )
{
	strcpy( out, in );
	if ( view->StringWidth(out) > size )
	{
		size -= view->StringWidth("...");
		while ( view->StringWidth(out) > size )
		{
			out[strlen(out)-1] = 0;
		}
		strcat(out, "...");
	}
}

basTumbnailView::basTumbnailView(BRect rect)
	: BView( rect, NULL, B_FOLLOW_ALL_SIDES, B_WILL_DRAW | B_FRAME_EVENTS | B_NAVIGABLE ),
	lastTumbnailsPerRow(0), tumbnailsPerRow(0),	movedTumbnail(0), movedTumbnailIndex(-1), 
	doingDraggingSelection(false), startedDoingDraggingSelection(false), RemoveInstedOfAdd(false)
{
	((basApplication*)be_app)->GetSettingInt( "basTumbnailView::TumbnailSize", &tumbnailSize );
	((basApplication*)be_app)->GetSettingBool( "basTumbnailView::useSmoothScaling", &useSmoothScaling );

	((basApplication*)be_app)->GetSettingInt( "basTumbnailView::leftButtonBehaviour", &leftButtonBehaviour );
	((basApplication*)be_app)->GetSettingInt( "basTumbnailView::middleButtonBehaviour", &middleButtonBehaviour );
	((basApplication*)be_app)->GetSettingInt( "basTumbnailView::rightButtonBehaviour", &rightButtonBehaviour );
	((basApplication*)be_app)->GetSettingInt( "basTumbnailView::doubleButtonBehaviour", &doubleButtonBehaviour );

	int temp;
	((basApplication*)be_app)->GetSettingInt( "basTumbnailView::TumbnailBackground", &temp);
	to_rgb( temp, TumbnailBackgroundColor);
	((basApplication*)be_app)->GetSettingInt( "basTumbnailView::TumbnailHover", &temp);
	to_rgb( temp, TumbnailHoverColor);
	rgb_color tmp;
	((basApplication*)be_app)->GetSettingInt( "basTumbnailView::ViewColor", &temp);
	to_rgb( temp, tmp);

	dragRectArea = BRect(-1, -1, -1, -1);		

	new basTumbnailCache();

	bool temp_bool;
	((basApplication*)be_app)->GetSettingBool( "basApplication::TumbnailCachingEnabled", &temp_bool );
	TumbnailCache->Enable(temp_bool);

	SetViewColor( BeListSelectGrey );

	SetLowColor( BeInactiveControlGrey );
	myContainer = new BetterScrollView( NULL, this, B_FOLLOW_ALL_SIDES, B_FRAME_EVENTS, false, true, true, B_FANCY_BORDER);
	myContainer->SetDataRect( BRect(0, 0, 0, 0), true);

	PopUp = new BPopUpMenu( "basTumbnailView::PopUp", false, false );
	PopUp->AddItem( new BMenuItem(Locale->Text(TEXT_SHOW_IMAGE, "Show Image"), new BMessage( B_MENU_SHOW_IMAGE )) );
	PopUp->AddItem( new BMenuItem(Locale->Text(TEXT_SHOW_IMAGE_IN_NEW_WINDOW, "Show Image in a new Window"), new BMessage( B_MENU_SHOW_IMAGE_IN_NEW_WINDOW )) );

	PopUp->AddSeparatorItem();
	PopUp->AddItem( new BMenuItem(Locale->Text(TEXT_OPEN_WITH_PREFERED_APPLICATION, "Open with prefered application"), new BMessage( B_MENU_OPEN_WITH_PREFERED_APPLICATION)));
	
	TumbnailSizePopUp = new BMenu( Locale->Text(TEXT_TUMBNAIL_SIZE, "Thumbnail Size" ));
	TumbnailSizePopUp->SetRadioMode(true);
	BMenuItem *item = new BMenuItem("25 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 25 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	item = new BMenuItem("50 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 50 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	item = new BMenuItem("75 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 75 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	item = new BMenuItem("100 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 100 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	item = new BMenuItem("150 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 150 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	item = new BMenuItem("200 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 200 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	PopUp->AddSeparatorItem();
	PopUp->AddItem( new BMenuItem(Locale->Text(TEXT_MOVE_IMAGE_TO_TRASH, "Move Image to Trash"), new BMessage( B_MENU_DELETE_IMAGE )) );
	PopUp->AddItem( new BMenuItem(Locale->Text(TEXT_MOVE_IMAGE_TO_DOTDOTDOT, "Move Image to ..."), new BMessage( B_MENU_MOVE_IMAGE )) );
	PopUp->AddItem( new BMenuItem(Locale->Text(TEXT_COPY_IMAGE_TO_DOTDOTDOT, "Copy Image to ..."), new BMessage( B_MENU_COPY_IMAGE )) );
	PopUp->AddSeparatorItem();
	PopUp->AddItem( new BMenuItem(Locale->Text(TEXT_ADD_THIS_IMAGE_TO_THE_IMAGELIST, "Add this image to the ImageList"), new BMessage( B_MENU_ADD_IMAGE_TO_IMAGELIST )) );
	PopUp->AddSeparatorItem();
	PopUp->AddItem( new BMenuItem(Locale->Text(TEXT_SEND_THIS_IMAGE_BY_EMAIL_DOTDOTDOT, "Send this image by E-Mail..."), new BMessage( B_MENU_SEND_IMAGE_BY_EMAIL )) );

	PopUp->AddSeparatorItem();
	PopUp->AddItem( TumbnailSizePopUp );

	msPopUp = new BPopUpMenu( "basTumbnailView::msPopUp", false, false );
	TumbnailSizePopUp = new BMenu( Locale->Text(TEXT_TUMBNAIL_SIZE, "Thumbnail Size" ));
	TumbnailSizePopUp->SetRadioMode(true);
	item = new BMenuItem("25 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 25 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	item = new BMenuItem("50 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 50 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	item = new BMenuItem("75 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 75 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	item = new BMenuItem("100 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 100 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	item = new BMenuItem("150 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 150 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	item = new BMenuItem("200 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 200 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
//	TODO:
//	msPopUp->AddSeparatorItem();
//	msPopUp->AddItem( new BMenuItem("Move Image to Trash", new BMessage( B_MENU_DELETE_IMAGE )) );
//	msPopUp->AddItem( new BMenuItem("Move Image to ...", new BMessage( B_MENU_MOVE_IMAGE )) );
//	msPopUp->AddItem( new BMenuItem("Copy Image to ...", new BMessage( B_MENU_COPY_IMAGE )) );
//	msPopUp->AddSeparatorItem();
	msPopUp->AddItem( new BMenuItem(Locale->Text(TEXT_ADD_SELECTED_IMAGES_TO_THE_IMAGELIST, "Add selected images to the ImageList"), new BMessage( B_MENU_ADD_IMAGE_TO_IMAGELIST )) );

	msPopUp->AddSeparatorItem();
	msPopUp->AddItem( TumbnailSizePopUp );

	nosPopUp = new BPopUpMenu( "basTumbnailView::nosPopUp", false, false );
	TumbnailSizePopUp = new BMenu( Locale->Text(TEXT_TUMBNAIL_SIZE, "Thumbnail Size" ));
	TumbnailSizePopUp->SetRadioMode(true);
	item = new BMenuItem("25 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 25 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	item = new BMenuItem("50 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 50 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	item = new BMenuItem("75 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 75 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	item = new BMenuItem("100 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 100 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	item = new BMenuItem("150 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 150 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	item = new BMenuItem("200 px", new BMessage( B_MENU_SET_TUMBNAIL_SIZE));
	if ( tumbnailSize == 200 )
		item->SetMarked(true);
	TumbnailSizePopUp->AddItem( item );
	nosPopUp->AddItem( new BMenuItem(Locale->Text(TEXT_ADD_ALL_IMAGES_TO_THE_IMAGELIST, "Add all images to the ImageList"), new BMessage(B_MENU_ADD_ALL_IMAGES_TO_IMAGELIST)));

	nosPopUp->AddSeparatorItem();
	nosPopUp->AddItem( TumbnailSizePopUp );

	StatusBar = new basStatusBar();
	StatusBar->SetLabel( "" );
	allContainer = new basStatusBarView( myContainer, StatusBar );

	tumbnailsSelection = new BList();
}

basTumbnailView::~basTumbnailView()
{
	delete tumbnailsSelection;
	CleanContents();
	delete TumbnailCache;
}

BView *basTumbnailView::Container()
{
	return allContainer;
}

void basTumbnailView::ShowFolderEaner( BEntry *ent, char *name, entry_ref &ref, node_ref &nref )
{
	BNode node(ent);
	BNodeInfo info(&node);
	char type[B_FILE_NAME_LENGTH];

	if ( ent->IsSymLink() )
	{
		BSymLink link( ent );
		char _path[B_PATH_NAME_LENGTH];
		if ( link.ReadLink( _path, B_PATH_NAME_LENGTH) > 0 )
		{
			if ( !link.IsAbsolute() )
			{
				BEntry pent;
				ent->GetParent( &pent );
				BPath pent_path( &pent );
				char tmp[B_PATH_NAME_LENGTH];
				sprintf(tmp, "%s/%s", pent_path.Path(), _path );
				strcpy(_path, tmp );
			}
			BEntry lent(_path);
			ShowFolderEaner(&lent, name, ref, nref);
		}
	}
	else
	{
		if ( info.GetType(type) != B_OK )
		{
			BVolume thisvol( ref.device );
			if ( thisvol.KnowsMime() )
			{
				BPath path(ent);
				if ( update_mime_info( path.Path(), 0, 1, 0) != B_OK )
					return;
				if ( info.GetType(type) != B_OK )
					return;
			}
			else
				return;
		}
		if ( strncasecmp(type, "image", 5) == 0 )
		{
			BPath path(ent);

			Tumbnail *tb = new Tumbnail();
			TumbnailList.AddItem(tb);
			char out[64];
			if (!name)
				name = (char *)path.Leaf();
			EnshortStringToFit( name, out, tumbnailSize, this);
			tb->Name = new char[strlen(out)+1];
			tb->realName = new char[ strlen(name)+1];
			strcpy(tb->realName, name);
			strcpy(tb->Name, out);
			tb->_fNameWidth = StringWidth(out);
			Entry_StringSize( ent, NULL, tb->sizeString );
			EnshortStringToFit( tb->sizeString, tb->sizeString, tumbnailSize, this );
			tb->_fsizeStringWidth = StringWidth(tb->sizeString);
			tb->Entry = ref;
			tb->Node = nref;

			if (!tb->Image)
			{
				int x, y;
				tb->Image = TumbnailCache->GetTumbnail( path.Path(), &x, &y, tumbnailSize );
				if ( tb->Image )
					sprintf(tb->bitmapSizeString, "%ix%i", x, y);
				else
					strcpy( tb->bitmapSizeString, "?x?" );
			}
			EnshortStringToFit( tb->bitmapSizeString, tb->bitmapSizeString, tumbnailSize, this );
			tb->_fbitmapSizeStringWidth = StringWidth(tb->bitmapSizeString);

			BMessage *msg = new BMessage(B_TUMBNAIL_VIEW_ADDED_ITEM);
			msg->AddInt32( "tumbnail_item_index", numberOfTumbnails-1 );
			Window()->PostMessage(msg);
		}
	}
}

void basTumbnailView::ShowFolderEaner( char *name, int size, BBitmap *_tb )
{
	Tumbnail *tb = new Tumbnail();
	TumbnailList.AddItem(tb);
	char out[64];
	EnshortStringToFit( name, out, tumbnailSize, this);
	tb->Name = new char[strlen(out)+1];
	tb->realName = new char[ strlen(name)+1];
	strcpy(tb->realName, name);
	strcpy(tb->Name, out);
	tb->_fNameWidth = StringWidth(out);
	Entry_StringSize( size, tb->sizeString );
	EnshortStringToFit( tb->sizeString, tb->sizeString, tumbnailSize, this );
	tb->_fsizeStringWidth = StringWidth(tb->sizeString);

	if ( _tb )
		tb->Image = new BBitmap(_tb);
	else
		tb->Image = _tb;

	strcpy( tb->bitmapSizeString, "?x?" );
	EnshortStringToFit( tb->bitmapSizeString, tb->bitmapSizeString, tumbnailSize, this );
	tb->_fbitmapSizeStringWidth = StringWidth(tb->bitmapSizeString);

	BMessage *msg = new BMessage(B_TUMBNAIL_VIEW_ADDED_ITEM);
	msg->AddInt32( "tumbnail_item_index", numberOfTumbnails-1 );
	Window()->PostMessage(msg);
}

void basTumbnailView::MouseDown( BPoint point )
{
	if ( !Window()->IsActive() )
		Window()->Activate(true);
	else
	{

		MakeFocus();

		BMessage* msg = Window()->CurrentMessage();
		int32 button = msg->FindInt32("buttons");
		int32 clicks = msg->FindInt32("clicks");

		int x_on_width = (int)point.x / (tumbnailSize+20);
		int y_on_height = (int)point.y / TUMBNAILSIZE28;
		int is = tumbnailsPerRow * y_on_height + x_on_width;
		register int m_i = (tumbnailSize+20)*x_on_width;
		register int m_j = TUMBNAILSIZE28*y_on_height;
		BRect retz( m_i+5, m_j+5, m_i+tumbnailSize+15, m_j+TUMBNAILSIZE28-3);

		if ( (x_on_width < tumbnailsPerRow) && retz.Contains(point) && (is < numberOfTumbnails) )
		{
			if ( (clicks >= 2 && ((button & B_PRIMARY_MOUSE_BUTTON) == B_PRIMARY_MOUSE_BUTTON)) && ( FirstClickTumbnail == TUMBNAILLIST(is) ))
			{
				FirstClickTumbnail = NULL;
				switch ( doubleButtonBehaviour )
				{
					case TBV_OPEN_IMAGE:
					{
						BEntry entry(&TUMBNAILLIST(is)->Entry);
						((basWindow*)Window())->DoSelectFunction( &entry );
						basFileListView *flv = ((basWindow*)Window())->FileListView;
						int i_s = flv->CountItems();
						basFileListItem **items = (basFileListItem**)flv->Items();
						for ( int i=0; i<i_s; i++)
						{
							if ( items[i]->Ref == TUMBNAILLIST(is)->Entry )
							{
								flv->Select( i );
								break;
							}
						}
					}
					break;
				}
			}
			else
			{
				FirstClickTumbnail = TUMBNAILLIST(is);

				if ( ((leftButtonBehaviour == TBV_OPEN_IMAGE) && (button & B_PRIMARY_MOUSE_BUTTON) == B_PRIMARY_MOUSE_BUTTON) ||
					 ((rightButtonBehaviour == TBV_OPEN_IMAGE) && (button & B_SECONDARY_MOUSE_BUTTON) == B_SECONDARY_MOUSE_BUTTON) ||
					 ((middleButtonBehaviour == TBV_OPEN_IMAGE) && (button & B_TERTIARY_MOUSE_BUTTON) == B_TERTIARY_MOUSE_BUTTON)
					 )
				{
					BEntry entry(&TUMBNAILLIST(is)->Entry);
					((basWindow*)Window())->DoSelectFunction( &entry );
					basFileListView *flv = ((basWindow*)Window())->FileListView;
					int i_s = flv->CountItems();
					basFileListItem **items = (basFileListItem**)flv->Items();
					for ( int i=0; i<i_s; i++)
					{
						if ( items[i]->Ref == TUMBNAILLIST(is)->Entry )
						{
							flv->Select( i );
							break;
						}
					}
				}
				else if ( ((rightButtonBehaviour == TBV_OPEN_MENU) && (button & B_SECONDARY_MOUSE_BUTTON) == B_SECONDARY_MOUSE_BUTTON) ||
						  ((leftButtonBehaviour == TBV_OPEN_MENU) && (button & B_PRIMARY_MOUSE_BUTTON) == B_PRIMARY_MOUSE_BUTTON) ||
						  ((middleButtonBehaviour == TBV_OPEN_MENU) && (button & B_TERTIARY_MOUSE_BUTTON) == B_TERTIARY_MOUSE_BUTTON)
							)
				{
					ConvertToScreen(&point);
					PopUpedTumbnail = TUMBNAILLIST(is);
					PopUpedTumbnailIndex = is;
					BMenuItem *sel;
					if ( tumbnailsSelection->CountItems() >= 2 )
						sel = msPopUp->Go( point );
					else
						sel = PopUp->Go( point );
					if ( sel )
					{
						if ( sel->Message() )
						{
							BMessage *msg = new BMessage ( sel->Message()->what );
							msg->AddInt32( "index", sel->Menu()->IndexOf( sel ) );
							Window()->PostMessage( msg );
						}
					}
				}
				else if ( ((rightButtonBehaviour == TBV_DRAG_DROP) && (button & B_SECONDARY_MOUSE_BUTTON) == B_SECONDARY_MOUSE_BUTTON) ||
						  ((leftButtonBehaviour == TBV_DRAG_DROP) && (button & B_PRIMARY_MOUSE_BUTTON) == B_PRIMARY_MOUSE_BUTTON) ||
						  ((middleButtonBehaviour == TBV_DRAG_DROP) && (button & B_TERTIARY_MOUSE_BUTTON) == B_TERTIARY_MOUSE_BUTTON)
							)
				{
					BMessage msg;
					msg.AddRef( "refs", &TUMBNAILLIST(is)->Entry );

					register int m_i = (tumbnailSize+20)*x_on_width;
					register int m_j = TUMBNAILSIZE28*y_on_height;
					DragMessage(&msg, BRect(m_i+10, m_j+10, m_i+tumbnailSize+10, m_j+tumbnailSize+10) );
				}
				else if ( ((rightButtonBehaviour == TBV_DO_SELECTION) && (button & B_SECONDARY_MOUSE_BUTTON) == B_SECONDARY_MOUSE_BUTTON) ||
						  ((leftButtonBehaviour == TBV_DO_SELECTION) && (button & B_PRIMARY_MOUSE_BUTTON) == B_PRIMARY_MOUSE_BUTTON) ||
						  ((middleButtonBehaviour == TBV_DO_SELECTION) && (button & B_TERTIARY_MOUSE_BUTTON) == B_TERTIARY_MOUSE_BUTTON)
							)
				{
					uint32 mod = modifiers();
					RemoveInstedOfAdd = mod & B_COMMAND_KEY;
					if ( !(mod & B_SHIFT_KEY) && !RemoveInstedOfAdd)
					{
						if ( !tumbnailsSelection->IsEmpty() )
						{
							tumbnailsSelection->MakeEmpty();
							Invalidate();
						}
					}
					if ( RemoveInstedOfAdd )
					{
						if ( tumbnailsSelection->HasItem(TUMBNAILLIST(is) ) )
							tumbnailsSelection->RemoveItem(TUMBNAILLIST(is));
					}
					else
					{
						if ( !tumbnailsSelection->HasItem(TUMBNAILLIST(is) ) )
							tumbnailsSelection->AddItem(TUMBNAILLIST(is));
					}
					Invalidate();
				}
			}
		}
		else
		{
			if ( ((rightButtonBehaviour == TBV_OPEN_MENU) && (button & B_SECONDARY_MOUSE_BUTTON) == B_SECONDARY_MOUSE_BUTTON) ||
				 ((leftButtonBehaviour == TBV_OPEN_MENU) && (button & B_PRIMARY_MOUSE_BUTTON) == B_PRIMARY_MOUSE_BUTTON) ||
				 ((middleButtonBehaviour == TBV_OPEN_MENU) && (button & B_TERTIARY_MOUSE_BUTTON) == B_TERTIARY_MOUSE_BUTTON))
			{
				ConvertToScreen(&point);
				BMenuItem *sel = nosPopUp->Go( point );
				if ( sel )
				{
					BMessage *msg = new BMessage ( sel->Message()->what );
					msg->AddInt32( "index", sel->Menu()->IndexOf( sel ) );
					Window()->PostMessage( msg );
				}
			}
			else if ( ((rightButtonBehaviour == TBV_DO_SELECTION) && (button & B_SECONDARY_MOUSE_BUTTON) == B_SECONDARY_MOUSE_BUTTON) ||
					  ((leftButtonBehaviour == TBV_DO_SELECTION) && (button & B_PRIMARY_MOUSE_BUTTON) == B_PRIMARY_MOUSE_BUTTON) ||
					  ((middleButtonBehaviour == TBV_DO_SELECTION) && (button & B_TERTIARY_MOUSE_BUTTON) == B_TERTIARY_MOUSE_BUTTON)
						)
			{
				uint32 mod = modifiers();
				RemoveInstedOfAdd = mod & B_COMMAND_KEY;
				if ( !(mod & B_SHIFT_KEY) && !RemoveInstedOfAdd)
				{
					if ( !tumbnailsSelection->IsEmpty() )
					{
						tumbnailsSelection->MakeEmpty();
						Invalidate();
					}
				}
				startedDoingDraggingSelection = true;
				doingDraggingSelectionPoint = point;
				SetMouseEventMask(B_POINTER_EVENTS, B_LOCK_WINDOW_FOCUS );
			}
		}
	}
}

void basTumbnailView::MouseUp( BPoint where )
{
	if ( doingDraggingSelection )
	{
		if ( ! (where == doingDraggingSelectionPoint) )
		{
			doingDraggingSelection = false;
			BRect area;
			BRegion update_area;

			if ( where.x < doingDraggingSelectionPoint.x )
			{
				if ( where.y < doingDraggingSelectionPoint.y )
					area = BRect( where, doingDraggingSelectionPoint );
				else
					area = BRect( where.x, doingDraggingSelectionPoint.y, doingDraggingSelectionPoint.x, where.y);
			}
			else
			{
				if ( where.y < doingDraggingSelectionPoint.y )
					area = BRect( doingDraggingSelectionPoint.x, where.y, where.x, doingDraggingSelectionPoint.y);
				else
					area = BRect( doingDraggingSelectionPoint, where );
			}

			bool do_invalidate = false;

			for ( int i=0; i<numberOfTumbnails; i++ )
			{
				int x_on_width = i%tumbnailsPerRow;
				int y_on_height = i/tumbnailsPerRow;
				register int m_i = (tumbnailSize+20)*x_on_width;
				register int m_j = TUMBNAILSIZE28*y_on_height;
				BRect retz( m_i+5, m_j+5, m_i+tumbnailSize+15, m_j+TUMBNAILSIZE28-3);
				if ( area.Intersects( retz) )
				{
					if ( RemoveInstedOfAdd )
					{
						if ( tumbnailsSelection->HasItem(TUMBNAILLIST(i) ) )
						{
							tumbnailsSelection->RemoveItem(TUMBNAILLIST(i));
							do_invalidate = true;
						}
					}
					else
					{
						if ( !tumbnailsSelection->HasItem(TUMBNAILLIST(i) ) )
						{
							tumbnailsSelection->AddItem(TUMBNAILLIST(i));
							do_invalidate = true;
						}
					}
					update_area.Include(retz);
				}
			}

			SetDrawingMode(B_OP_INVERT);
			StrokeRect(dragRectArea);
			SetDrawingMode(B_OP_OVER);

			if ( do_invalidate )
			{
				BRect update_rect( update_area.Frame() );
				update_rect.left --;
				update_rect.top --;
				update_rect.right += 2;
				update_rect.bottom += 2;
				Invalidate(update_rect);
			}

			dragRectArea = BRect(-1, -1, -1, -1);
		}
	}
}

void basTumbnailView::MouseMoved( BPoint point, uint32 code, const BMessage *msg )
{
/*	if (!msg)
	{
		int x_on_width = (int)point.x / (tumbnailSize+20);
		int y_on_height = (int)point.y / TUMBNAILSIZE28;
		int was = movedTumbnailIndex;
		int is = tumbnailsPerRow * y_on_height + x_on_width;
		if ( (x_on_width < tumbnailsPerRow ) && (is < numberOfTumbnails) )
		{
			if ( is != movedTumbnailIndex )
			{
				movedTumbnailIndex = is;
				movedTumbnail = TumbnailList[is];
				int row = was / tumbnailsPerRow;
				int col = was % tumbnailsPerRow;
				BRect update_rect( x_on_width * (tumbnailSize+20), y_on_height * TUMBNAILSIZE28, (x_on_width+1) *(tumbnailSize+20), (y_on_height+1) *TUMBNAILSIZE28 + 10);
				Invalidate(update_rect);
				BRect update_rect2( col * (tumbnailSize+20), row * TUMBNAILSIZE28, (col+1) *(tumbnailSize+20), (row+1) *TUMBNAILSIZE28 + 10);
				Invalidate(update_rect2);
			}
		}
		else
		{
			if ( movedTumbnailIndex >= 0 )
			{
				movedTumbnail = NULL;
				movedTumbnailIndex = -1;
				int row = was / tumbnailsPerRow;
				int col = was % tumbnailsPerRow;
				BRect update_rect2( col * (tumbnailSize+20), row * TUMBNAILSIZE28, (col+1) *(tumbnailSize+20), (row+1) *TUMBNAILSIZE28 + 10);
				Invalidate(update_rect2);
			}
		}
	}*/
	if ( doingDraggingSelection )
	{
//		TODO:
//		if ( point.y < 0 )
//		{
//			ScrollBy( 0, point.y / 10);
//		}
//		else if ( point.y >= Bounds().Height() )
//		{
//			ScrollBy( 0, (point.y - Bounds().Height()) / 10.0 );
//		}

		SetHighColor(Black);

		SetDrawingMode(B_OP_INVERT);
		StrokeRect(dragRectArea);

		if ( point.x < doingDraggingSelectionPoint.x )
		{
			if ( point.y < doingDraggingSelectionPoint.y )
				dragRectArea = BRect( point, doingDraggingSelectionPoint );
			else
				dragRectArea = BRect( point.x, doingDraggingSelectionPoint.y, doingDraggingSelectionPoint.x, point.y);
		}
		else
		{
			if ( point.y < doingDraggingSelectionPoint.y )
				dragRectArea = BRect( doingDraggingSelectionPoint.x, point.y, point.x, doingDraggingSelectionPoint.y);
			else
				dragRectArea = BRect( doingDraggingSelectionPoint, point );
		}

		StrokeRect(dragRectArea);

		SetDrawingMode(B_OP_OVER);
	}

	if ( startedDoingDraggingSelection )
	{
		BPoint point;
		uint32 button;
		GetMouse( &point, &button );
		if ( ((rightButtonBehaviour == TBV_DO_SELECTION) && (button & B_SECONDARY_MOUSE_BUTTON) == B_SECONDARY_MOUSE_BUTTON) ||
			  ((leftButtonBehaviour == TBV_DO_SELECTION) && (button & B_PRIMARY_MOUSE_BUTTON) == B_PRIMARY_MOUSE_BUTTON) ||
			  ((middleButtonBehaviour == TBV_DO_SELECTION) && (button & B_TERTIARY_MOUSE_BUTTON) == B_TERTIARY_MOUSE_BUTTON) )
		{
			doingDraggingSelection = true;
		}
		startedDoingDraggingSelection = false;
	}
}

void basTumbnailView::FrameResized( float x, float y)
{
	tumbnailsPerRow = (int)x/(tumbnailSize+20);
	UpdateScrollbars();
	if ( tumbnailsPerRow != lastTumbnailsPerRow )
	{
		lastTumbnailsPerRow = tumbnailsPerRow;
		Invalidate();
	}
}

void basTumbnailView::UpdateScrollbars()
{
	if ( tumbnailsPerRow && numberOfTumbnails )
	{
		float data = TUMBNAILSIZE28*((numberOfTumbnails/tumbnailsPerRow)+((numberOfTumbnails%tumbnailsPerRow)?1:0));
		myContainer->SetDataRect(BRect(0, 0, 0, data<Bounds().Height()?0:data));
	}
	else
		myContainer->SetDataRect( BRect(0, 0, 0, 0));
}

void basTumbnailView::Draw( BRect draw_rect )
{
	BRegion Region;
	Region.Include(draw_rect);
	ConstrainClippingRegion(&Region);
	if ( tumbnailsPerRow )
	{
		int k = 0, n_rows = numberOfTumbnails / tumbnailsPerRow + ((numberOfTumbnails % tumbnailsPerRow)?1:0);
		for ( int j=0; j<n_rows; j++ )
		{
			for ( int i=0; i<tumbnailsPerRow; i++)
			{
				register int m_i = (tumbnailSize+20)*i;
				register int m_j = TUMBNAILSIZE28*j;
				Tumbnail *tb = TUMBNAILLIST(j*tumbnailsPerRow+i);
				if ( tb == movedTumbnail )
				{
					SetLowColor( TumbnailHoverColor );
					SetHighColor( TumbnailHoverColor );
					}
				else
				{
//					CHANGED:
//					SetLowColor(TumbnailBackgroundColor);
//					SetHighColor(TumbnailBackgroundColor);
					SetLowColor((rgb_color){222, 222, 222, 255});
					SetHighColor((rgb_color){222, 222, 222, 255});
				}

				FillRect( BRect( m_i+5, m_j+5, m_i+tumbnailSize+15, m_j+TUMBNAILSIZE28-3));

				SetHighColor(BeDarkShadow);
				FillRect( BRect(m_i+5, m_j+5, m_i+tumbnailSize+15, m_j+5) );
				FillRect( BRect(m_i+5, m_j+5, m_i+5, m_j+TUMBNAILSIZE28-3) );
				FillRect( BRect(m_i+tumbnailSize+15, m_j+5, m_i+tumbnailSize+15, m_j+TUMBNAILSIZE28-3) );
				FillRect( BRect(m_i+5, m_j+TUMBNAILSIZE28-3, m_i+tumbnailSize+15, m_j+TUMBNAILSIZE28-3) );
				SetHighColor(White);
				FillRect( BRect(m_i+6, m_j+6, m_i+tumbnailSize+12, m_j+6) );
				FillRect( BRect(m_i+6, m_j+6, m_i+6, m_j+TUMBNAILSIZE28-4) );

				SetHighColor(BeShadow);
				FillRect( BRect(m_i+tumbnailSize+14, m_j+6, m_i+tumbnailSize+14, m_j+TUMBNAILSIZE28-4) );
				FillRect( BRect(m_i+7, m_j+TUMBNAILSIZE28-4, m_i+tumbnailSize+14, m_j+TUMBNAILSIZE28-4) );

				if ( tumbnailsSelection->HasItem(tb) )
				{
					SetHighColor(BeFocusBlue);
					StrokeRect( BRect( m_i+4, m_j+4, m_i+tumbnailSize+16, m_j+TUMBNAILSIZE28-2));
				}

				BBitmap *bb = tb->Image;
				if ( bb )
				{
					float fw = bb->Bounds().Width();
					float fh = bb->Bounds().Height();
					float _w = m_i + 10 + ((tumbnailSize-fw)/2);
					float _h = m_j + 10 + ((tumbnailSize-fh)/2);
					DrawBitmap( bb, BRect(_w, _h, _w+fw, _h+fh) );
				}

				SetHighColor(Black);
				if ( tb->Name )
					DrawString( tb->Name, BPoint(m_i+10+((tumbnailSize-tb->_fNameWidth)/2), m_j+tumbnailSize+21));
				if ( tb->sizeString )
					DrawString( tb->sizeString, BPoint(m_i+10+((tumbnailSize-tb->_fsizeStringWidth)/2), m_j+tumbnailSize+21+12));
				if ( tb->bitmapSizeString )
					DrawString( tb->bitmapSizeString, BPoint(m_i+10+((tumbnailSize-tb->_fbitmapSizeStringWidth)/2), m_j+tumbnailSize+21+24));
				k++;
				if ( k >= numberOfTumbnails )
					break;
			}
			if ( k >= numberOfTumbnails )
				break;
		}
	}
	ConstrainClippingRegion(NULL);
}

void basTumbnailView::CleanContents()
{
	int i;
	for ( i=0; i<numberOfTumbnails; i++)
	{
		delete (Tumbnail*)(TumbnailList.RemoveItem(i));
	}
	TumbnailList = 0;
	Invalidate();
}

void basTumbnailView::AddedItem( int index)
{
	if ( tumbnailsPerRow )
	{
		UpdateScrollbars();
		int row = index / tumbnailsPerRow;
		int col = index % tumbnailsPerRow;
		BRect update_rect( col * (tumbnailSize+20), row * TUMBNAILSIZE28, (col+1) *(tumbnailSize+20), (row+1) *TUMBNAILSIZE28 + 10);
		Invalidate(update_rect);
		char num[128];
		sprintf(num, Locale->Text(TEXT_I_IMAGES, "%i images"), index+1 );
		StatusBar->SetLabel(num);
	}
}

void basTumbnailView::SetTumbnailSize( int size )
{
	tumbnailSize = size;
	((basApplication*)be_app)->SetSettingInt( "basTumbnailView::TumbnailSize", tumbnailSize);
	if ( Window() )
	{
		Window()->Lock();
		FrameResized( Bounds().Width(), Bounds().Height() );
		Window()->Unlock();
	}
}

void basTumbnailView::MessageReceived( BMessage *msg )
{
	switch ( msg->what )
	{
	default:
		BView::MessageReceived(msg);
	}
}

void basTumbnailView::RemovedTumbnail( int v )
{
	Window()->Lock();
	Tumbnail *tb = (Tumbnail*)(TumbnailList.RemoveItem(v));
	if (tb)
		delete tb;
	FrameResized( Bounds().Width(), Bounds().Height() );
	Invalidate();
	Window()->Unlock();
}

int basTumbnailView::FindTumbnailByNode( node_ref &nref )
{
	for ( int i = 0; i< numberOfTumbnails; i++ )
	{
		if ( TUMBNAILLIST(i)->Node == nref )
			return i;
	}
	return -1;
}

void basTumbnailView::UpdateTumbnail(int index, entry_ref *eref, char *name )
{
	Tumbnail *tb = TUMBNAILLIST(index);
	BEntry ent;
	if ( eref )
	{
		tb->Entry = *eref;
		ent.SetTo(eref);
	}
	else
		ent.SetTo(&tb->Entry);
	if ( name )
	{
		delete tb->realName;
		tb->realName = new char[ strlen(eref->name) + 1];
		strcpy( tb->realName, eref->name );
		char out[64];
		EnshortStringToFit( name, out, tumbnailSize, this);
		tb->Name = new char[strlen(out)+1];
		strcpy(tb->Name, out);
		tb->_fNameWidth = StringWidth(out);
	}
	Entry_StringSize( &ent, NULL, tb->sizeString );
	EnshortStringToFit( tb->sizeString, tb->sizeString, tumbnailSize, this );
	tb->_fsizeStringWidth = StringWidth(tb->sizeString);
	int row = index / tumbnailsPerRow;
	int col = index % tumbnailsPerRow;
	BRect update_rect( col * (tumbnailSize+20), row * TUMBNAILSIZE28, (col+1) *(tumbnailSize+20), (row+1) *TUMBNAILSIZE28 + 10);
	Window()->Lock();
	Invalidate(update_rect);
	Window()->Unlock();
}

void basTumbnailView::SetLeftMouseButtonBehaviour( int i )
{
	leftButtonBehaviour = i;
}

void basTumbnailView::SetMiddleMouseButtonBehaviour( int i )
{
	middleButtonBehaviour = i;
}

void basTumbnailView::SetRightMouseButtonBehaviour( int i )
{
	rightButtonBehaviour = i;
}

void basTumbnailView::SetDoubleMouseButtonBehaviour( int i )
{
	doubleButtonBehaviour = i;
}

void basTumbnailView::SetTumbnailBackgroundColor(rgb_color C)
{
	memcpy( &TumbnailBackgroundColor, &C, sizeof(rgb_color));
}

void basTumbnailView::SetTumbnailHoverColor(rgb_color C)
{
	memcpy( &TumbnailHoverColor, &C, sizeof(rgb_color));
}

void basTumbnailView::SetStatusLabel( char *text )
{
	StatusBar->SetLabel(text);
}